@@ -5,8 +5,8 @@ test_models = [
     # "esm1_t34_670M_UR100",
     "esm1_t6_43M_UR50S",
     "esm1b_t33_650M_UR50S",
-    # "protbert",
-    "protbert_bfd",
+    "protbert",
+    # "protbert_bfd",
 ]
 
 
