@@ -4,19 +4,18 @@ import pytest
 test_sequences = ["AAAA", "AKKF", "AHHFK", "KKKKKKKLLL"]
 
 test_params = [
-    (1, list("ACDEFGHIKLMNPQRSTVWY"), "forward"),
-    (2, ["A", "F", "K"], "masked"),
-    (10, ["A"], "forward"),
+    (1, "forward"),
+    (2, "masked"),
+    (10, "forward"),
 ]
 
 
-@pytest.mark.parametrize("batch_size, tokens_list, pass_mode", test_params)
-def test_accuracy_type_and_range(init_model, batch_size, tokens_list, pass_mode):
+@pytest.mark.parametrize("batch_size, pass_mode", test_params)
+def test_accuracy_type_and_range(init_model, batch_size, pass_mode):
     test_trans = init_model
     accuracy = test_trans.compute_accuracy(
         test_sequences,
         batch_size=batch_size,
-        tokens_list=tokens_list,
         pass_mode=pass_mode,
     )
     assert isinstance(accuracy, float)
