@@ -4,24 +4,21 @@ specific to a given transformers implementation can inherit.
 It allows to derive probabilities, embeddings and log-likelihoods based on inputs
 sequences, and displays some properties of the transformer model.
 """
+import math
 import os
-from abc import ABC, abstractmethod
 from copy import deepcopy
 from os.path import join
 from pathlib import Path
-from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple, Union
+from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple, Type, Union
 
 import numpy as np
 import pytorch_lightning as pl
+import ray
 import torch
 import torch.tensor
-from biotransformers.lightning_utils.data import (
-    AlphabetDataLoader,
-    convert_ckpt_to_statedict,
-)
 from biotransformers.utils.constant import NATURAL_AAS_LIST
-from biotransformers.utils.gpus_utils import set_device
 from biotransformers.utils.logger import logger  # noqa
+from biotransformers.utils.tqdm_utils import ProgressBar
 from biotransformers.utils.utils import (
     _check_memory_embeddings,
     _check_memory_logits,
@@ -30,13 +27,10 @@ from biotransformers.utils.utils import (
     get_logs_version,
     load_fasta,
 )
-
-# from esm.data import read_fasta
+from biotransformers.wrappers.language_model import LanguageModel
 from pytorch_lightning import Trainer
 from pytorch_lightning.callbacks import ModelCheckpoint
 from pytorch_lightning.loggers import CSVLogger
-from torch.nn import DataParallel
-from tqdm import tqdm
 
 from ..lightning_utils.data import BioDataModule
 from ..lightning_utils.models import LightningModule
@@ -44,7 +38,7 @@ from ..lightning_utils.models import LightningModule
 log = logger("transformers_wrapper")
 
 
-class TransformersWrapper(ABC):
+class TransformersWrapper:
     """
     Abstract class that uses pretrained transformers model to evaluate
     a protein likelihood so as other insights.
@@ -53,107 +47,49 @@ class TransformersWrapper(ABC):
     def __init__(
         self,
         model_dir: str,
-        _device: Optional[str] = None,
-        multi_gpu: bool = False,
-        mask_bool: bool = False,
+        language_model_cls: Type[LanguageModel],
+        num_gpus: int = 0,
     ):
         """Initialize Transformers wrapper
 
+        TODO : os.environ["CUDA_VISIBLE_DEVICES"]="0" or export CUDA_VISIBLE_DEVICES="0,1"
+
         Args:
             model_dir: name directory of the pretrained model
-            _device: type of device to use (cpu or cuda).
-            multi_gpu: turn on to True to use multigpu
-            mask_bool: Wether to use mask or not for inference.
+            num_gpus: number of gpus to use. If set to 0, it uses the cpu.
         """
-        _device, _multi_gpu = set_device(_device, multi_gpu)
-
-        self._device = torch.device(_device)
-        self.multi_gpu = _multi_gpu
-        self.model_dir = model_dir
-        self.mask_bool = mask_bool
-
-    @property
-    def model_id(self) -> str:
-        """Model ID, as specified in the model directory"""
-        return self.model_dir.lower()
-
-    @property
-    @abstractmethod
-    def clean_model_id(self) -> str:
-        """Clean model ID (in case the model directory is not)"""
-
-    @property
-    @abstractmethod
-    def model_vocabulary(self) -> List[str]:
-        """Returns the whole vocabulary list"""
-
-    @property
-    @abstractmethod
-    def vocab_size(self) -> int:
-        """Returns the whole vocabulary size"""
-
-    @property
-    @abstractmethod
-    def mask_token(self) -> str:
-        """Representation of the mask token (as a string)"""
-
-    @property
-    @abstractmethod
-    def pad_token(self) -> str:
-        """Representation of the pad token (as a string)"""
-
-    @property
-    @abstractmethod
-    def begin_token(self) -> str:
-        """Representation of the beginning of sentence token (as a string).
-        Returns an empty string if no such token"""
-
-    @property
-    @abstractmethod
-    def end_token(self) -> str:
-        """Representation of the end of sentence token (as a string).
-        Returns an empty string if no such token."""
-
-    @property
-    @abstractmethod
-    def does_end_token_exist(self) -> bool:
-        """Returns true if a end of sequence token exists"""
-
-    @property
-    @abstractmethod
-    def token_to_id(self):
-        """Returns a function which maps tokens to IDs"""
-
-    @property
-    @abstractmethod
-    def embeddings_size(self) -> int:
-        """Returns size of the embeddings"""
+        self._model_dir = model_dir
+        self._num_gpus = num_gpus
+        if num_gpus >= 1:
+            assert torch.cuda.is_available(), "Cuda is not available."
+            assert torch.cuda.device_count() >= num_gpus, "Not enough available GPUs."
+        if num_gpus <= 1:
+            device = "cpu" if num_gpus == 0 else "cuda"
+            self._language_model = language_model_cls(model_dir=model_dir, device=device)
+            self._multi_gpus = False
+        else:
+            self._language_model = language_model_cls(model_dir=model_dir, device="cpu")
+            self._ray_cls = ray.remote(num_cpus=4, num_gpus=1)(language_model_cls)
+            self._workers = [
+                self._ray_cls.remote(model_dir=model_dir, device="cuda:0") for _ in range(num_gpus)
+            ]
+            self._multi_gpus = True
 
     def get_vocabulary_mask(self, tokens_list: List[str]) -> np.ndarray:
         """Returns a mask ove the model tokens."""
         # Compute a mask over vocabulary to decide which value to keep or not
         vocabulary_mask = np.array(
-            [1.0 if token in tokens_list else 0.0 for token in self.model_vocabulary]
+            [
+                1.0 if token in tokens_list else 0.0
+                for token in self._language_model.model_vocabulary
+            ]
         )
         return vocabulary_mask
 
-    @abstractmethod
-    def _process_sequences_and_tokens(
-        self, sequences_list: List[str]
-    ) -> Dict[str, torch.Tensor]:
-        """Function to transform tokens string to IDs; it depends on the model used"""
-
-    @abstractmethod
-    def _model_pass(
-        self, model_inputs: Dict[str, torch.tensor]
-    ) -> Tuple[torch.tensor, torch.tensor]:
-        """Function which computes logits and embeddings based on a list of sequences"""
-
-    @abstractmethod
-    def _get_alphabet_dataloader(self) -> AlphabetDataLoader:
-        """Function to build custom alphanbet"""
-
     def _get_num_batch_iter(self, model_inputs: Dict[str, Any], batch_size: int) -> int:
+        """
+        Get the number of batches when spliting model_inputs into chunks of size batch_size.
+        """
         num_of_sequences = model_inputs["input_ids"].shape[0]
         num_batch_iter = int(np.ceil(num_of_sequences / batch_size))
         return num_batch_iter
@@ -191,10 +127,13 @@ class TransformersWrapper(ABC):
             model_inputs["token_type_ids"],
         ):
             mask_id = []
-            for i in range(1, sum(binary_mask) - self.does_end_token_exist * 1):
+            mask_token = self._language_model.mask_token
+            does_end_token_exist = self._language_model.does_end_token_exist
+
+            for i in range(1, sum(binary_mask) - does_end_token_exist * 1):
                 mask_sequence = torch.tensor(
                     sequence[:i].tolist()
-                    + [self.token_to_id(self.mask_token)]
+                    + [self._language_model.token_to_id(mask_token)]
                     + sequence[i + 1 :].tolist(),
                     dtype=torch.int64,
                 )
@@ -224,9 +163,7 @@ class TransformersWrapper(ABC):
             model_outputs (torch.Tensor): shape -> (num_seqs, max_seq_len, vocab_size)
         """
         max_length = model_outputs.shape[1]
-        inf_tensor = -float("Inf") * torch.ones(
-            [1, model_outputs.shape[2]], dtype=torch.float32
-        )
+        inf_tensor = -float("Inf") * torch.ones([1, model_outputs.shape[2]], dtype=torch.float32)
         sequences_list = []
         start_id = 0
         for mask_id in masked_ids_list:
@@ -262,28 +199,37 @@ class TransformersWrapper(ABC):
                     * embeddings [num_seqs, max_len_seqs+1, embedding_size]
         """
         silent = kwargs.get("silent", False)
-        # Initialize logits and embeddings before looping over batches
-        logits = torch.Tensor()  # [num_seqs, max_len_seqs+1, vocab_size]
-        embeddings = torch.Tensor()  # [num_seqs, max_len_seqs+1, embedding_size]
-
-        for batch_inputs in tqdm(
-            self._generate_chunks(model_inputs, batch_size),
-            total=self._get_num_batch_iter(model_inputs, batch_size),
-            disable=silent,
-        ):
-            batch_logits, batch_embeddings = self._model_pass(batch_inputs)
-
-            embeddings = torch.cat((embeddings, batch_embeddings), dim=0)
-            logits = torch.cat((logits, batch_logits), dim=0)
+        num_inputs = model_inputs["input_ids"].shape[0]
+        n_updates = math.ceil(num_inputs / batch_size)
+        if self._multi_gpus:
+            jobs = []
+            # Initialize logits and embeddings before looping over batches
+            logits = torch.Tensor()  # [num_seqs, max_len_seqs+1, vocab_size]
+            embeddings = torch.Tensor()  # [num_seqs, max_len_seqs+1, embedding_size]
+            pb = ProgressBar(n_updates)
+            actor = pb.actor
+            for i, batch_inputs in enumerate(
+                self._generate_chunks(model_inputs, math.ceil(num_inputs / self._num_gpus))
+            ):
+                # Split large batch into smaller batches, when per GPU worker
+                # Send tqdm progress bar
+                jobs.append(
+                    self._workers[i].model_pass.remote(batch_inputs, batch_size, silent, actor)
+                )
+            pb.print_until_done()
+            # Launch parallel execution in background
+            outs = ray.get(jobs)
+            # Gather the result
+            for batch_logits, batch_embeddings in outs:
+                embeddings = torch.cat((embeddings, batch_embeddings), dim=0)
+                logits = torch.cat((logits, batch_logits), dim=0)
+        else:
+            logits, embeddings = self._language_model.model_pass(model_inputs, batch_size, silent)
 
         return logits, embeddings
 
     def _compute_logits(
-        self,
-        model_inputs: Dict[str, torch.Tensor],
-        batch_size: int,
-        pass_mode: str,
-        **kwargs
+        self, model_inputs: Dict[str, torch.Tensor], batch_size: int, pass_mode: str, **kwargs
     ) -> torch.Tensor:
         """Intermediate function to compute logits
 
@@ -297,14 +243,10 @@ class TransformersWrapper(ABC):
         """
         if pass_mode == "masked":
             model_inputs, masked_ids_list = self._repeat_and_mask_inputs(model_inputs)
-            logits, _ = self._model_evaluation(
-                model_inputs, batch_size=batch_size, **kwargs
-            )
+            logits, _ = self._model_evaluation(model_inputs, batch_size=batch_size, **kwargs)
             logits = self._gather_masked_outputs(logits, masked_ids_list)
         elif pass_mode == "forward":
-            logits, _ = self._model_evaluation(
-                model_inputs, batch_size=batch_size, **kwargs
-            )
+            logits, _ = self._model_evaluation(model_inputs, batch_size=batch_size, **kwargs)
         return logits
 
     def compute_logits(
@@ -330,26 +272,21 @@ class TransformersWrapper(ABC):
 
         if isinstance(sequences, str):
             sequences = load_fasta(sequences)
-
-        _check_sequence(sequences, self.model_dir, 1024)
-        _check_memory_logits(sequences, self.vocab_size, pass_mode)
+        _check_sequence(sequences, self._model_dir, 1024)
+        _check_memory_logits(sequences, self._language_model.vocab_size, pass_mode)
 
         # Perform inference in model to compute the logits
-        inputs = self._process_sequences_and_tokens(sequences)
+        inputs = self._language_model.process_sequences_and_tokens(sequences)
         labels = torch.unsqueeze(inputs["input_ids"], dim=-1)
         logits = self._compute_logits(inputs, batch_size, pass_mode, silent=silent)
-
         # Remove padded logits
         lengths = [len(sequence) for sequence in sequences]
         logits = [logit[:length, :] for logit, length in zip(list(logits), lengths)]
         labels = [label[:length, :] for label, length in zip(list(labels), lengths)]
-
         # Keep only corresponding to amino acids that are in the sequence
         logits = [
-            torch.gather(logit, dim=-1, index=label).numpy()
-            for logit, label in zip(logits, labels)
+            torch.gather(logit, dim=-1, index=label).numpy() for logit, label in zip(logits, labels)
         ]
-
         # List of arrays of shape (seq_length, 1)
         return logits
 
@@ -392,22 +329,18 @@ class TransformersWrapper(ABC):
         if isinstance(sequences, str):
             sequences = load_fasta(sequences)
         tokens_list = NATURAL_AAS_LIST if tokens_list is None else tokens_list
-
-        _check_sequence(sequences, self.model_dir, 1024)
-        _check_memory_logits(sequences, self.vocab_size, pass_mode)
+        _check_sequence(sequences, self._model_dir, 1024)
+        _check_memory_logits(sequences, self._language_model.vocab_size, pass_mode)
         _check_tokens_list(sequences, tokens_list)
 
         # Perform inference in model to compute the logits
-        inputs = self._process_sequences_and_tokens(sequences)
+        inputs = self._language_model.process_sequences_and_tokens(sequences)
         logits = self._compute_logits(inputs, batch_size, pass_mode, silent=silent)
-
         # Remove padded logits
         lengths = [len(sequence) for sequence in sequences]
         logits = [logit[:length, :] for logit, length in zip(list(logits), lengths)]
-
         # Set to -inf logits that correspond to tokens that are not in tokens list
         vocabulary_mask = torch.from_numpy(self.get_vocabulary_mask(tokens_list))
-
         # Avoid printing warnings
         np.seterr(divide="ignore")
 
@@ -417,7 +350,6 @@ class TransformersWrapper(ABC):
                 np.tile(np.log(vocabulary_mask), (logit.shape[0], 1))
             )
             masked_logits.append(masked_logit)
-
         # Use softmax to compute probabilities frm logits
         # Due to the -inf, probs of tokens that are not in token list will be zero
         softmax = torch.nn.Softmax(dim=-1)
@@ -426,15 +358,12 @@ class TransformersWrapper(ABC):
         def _get_probabilities_dict(probs: torch.Tensor) -> Dict[str, float]:
             return {
                 token: float(probs[i].cpu().numpy())
-                for i, token in enumerate(self.model_vocabulary)
+                for i, token in enumerate(self._language_model.model_vocabulary)
                 if token in tokens_list
             }
 
         probabilities_dict = [
-            {
-                key: _get_probabilities_dict(value)
-                for key, value in dict(enumerate(probs)).items()
-            }
+            {key: _get_probabilities_dict(value) for key, value in dict(enumerate(probs)).items()}
             for probs in probabilities
         ]
         return probabilities_dict
@@ -473,13 +402,9 @@ class TransformersWrapper(ABC):
         log_likelihoods = []
         for sequence, probabilities_dict in zip(sequences, probabilities):
             log_likelihood = np.sum(
-                [
-                    np.log(probabilities_dict[i][sequence[i]])
-                    for i in range(len(sequence))
-                ]
+                [np.log(probabilities_dict[i][sequence[i]]) for i in range(len(sequence))]
             )
             log_likelihoods.append(float(log_likelihood))
-
         return log_likelihoods
 
     def compute_embeddings(
@@ -513,31 +438,23 @@ class TransformersWrapper(ABC):
         Returns:
              Dict[str, List[np.ndarray]]: dict matching pool-mode and list of embeddings
         """
-
         if isinstance(sequences, str):
             sequences = load_fasta(sequences)
-
-        _check_sequence(sequences, self.model_dir, 1024)
-        _check_memory_embeddings(sequences, self.embeddings_size, pool_mode)
+        _check_sequence(sequences, self._model_dir, 1024)
+        _check_memory_embeddings(sequences, self._language_model.embeddings_size, pool_mode)
 
         # Get the sequences lengths
         lengths = [len(sequence) for sequence in sequences]
-
         # Compute a forward pass to get the embeddings
-        inputs = self._process_sequences_and_tokens(sequences)
-        _, embeddings = self._model_evaluation(
-            inputs, batch_size=batch_size, silent=silent
-        )
+        inputs = self._language_model.process_sequences_and_tokens(sequences)
+        _, embeddings = self._model_evaluation(inputs, batch_size=batch_size, silent=silent)
         embeddings = [emb.cpu().numpy() for emb in embeddings]
-
         # Remove class token and padding
         filtered_embeddings = [
             emb[1 : (length + 1), :] for emb, length in zip(list(embeddings), lengths)
         ]
-
         # Keep class token only
         cls_embeddings = [emb[0, :] for emb in list(embeddings)]
-
         embeddings_dict = {}
         # Keep only what's necessary
         if "full" in pool_mode:
@@ -545,10 +462,7 @@ class TransformersWrapper(ABC):
         if "cls" in pool_mode:
             embeddings_dict["cls"] = np.stack(cls_embeddings)
         if "mean" in pool_mode:
-            embeddings_dict["mean"] = np.stack(
-                [np.mean(e, axis=0) for e in filtered_embeddings]
-            )
-
+            embeddings_dict["mean"] = np.stack([np.mean(e, axis=0) for e in filtered_embeddings])
         return embeddings_dict
 
     def compute_accuracy(
@@ -571,21 +485,19 @@ class TransformersWrapper(ABC):
         """
         if isinstance(sequences, str):
             sequences = load_fasta(sequences)
-
-        _check_sequence(sequences, self.model_dir, 1024)
-        _check_memory_logits(sequences, self.vocab_size, pass_mode)
+        _check_sequence(sequences, self._model_dir, 1024)
+        _check_memory_logits(sequences, self._language_model.vocab_size, pass_mode)
 
         # Perform inference in model to compute the logits
-        inputs = self._process_sequences_and_tokens(sequences)
+        inputs = self._language_model.process_sequences_and_tokens(sequences)
         logits = self._compute_logits(inputs, batch_size, pass_mode, silent=silent)
+        # Get length of sequence
         labels = inputs["input_ids"]
 
         # Get the predicted labels
         predicted_labels = torch.argmax(logits, dim=-1)
-
         # Compute the accuracy
         accuracy = float(torch.mean(torch.eq(predicted_labels, labels).float()))
-
         return accuracy
 
     def load_model(self, model_dir: str, map_location=None):
@@ -601,29 +513,21 @@ class TransformersWrapper(ABC):
         if not os.path.isfile(model_dir):
             raise FileNotFoundError
 
-        if model_dir.endswith(".pt"):
-            load_model = torch.load(model_dir)
-            log.info("Load model %s" % model_dir)
-        elif model_dir.endswith(".ckpt"):
-            load_model = convert_ckpt_to_statedict(torch.load(model_dir)["state_dict"])
-            log.info("Load checkpoint %s" % model_dir)
+        if self._multi_gpus:
+            ray.get(
+                [worker._load_model.remote(model_dir, map_location) for worker in self._workers]
+            )
+            pass
         else:
-            raise ValueError("Expecting a .pt or .ckpt file")
-
-        if self.multi_gpu:
-            self.model.module.load_state_dict(load_model, map_location)  # type: ignore
-        else:
-            self.model.load_state_dict(load_model, map_location)  # type: ignore
-            self.model.eval()  # type: ignore
-
-    def save_model(self, exp_path: str, lightning_model: pl.LightningModule):
-        """Save pytorch model in logs directory
+            self._language_model._load_model(model_dir, map_location)  # type: ignore
 
+    def _save_model(self, exp_path: str, lightning_model: pl.LightningModule):
+        """Save pytorch model in pytorch-lightning logs directory
         Args:
             exp_path (str): path of the experiments directory in the logs
         """
         version = get_logs_version(exp_path)
-        model_dir = self.model_dir.replace("/", "_")
+        model_dir = self._model_dir.replace("/", "_")
         if version is not None:
             save_name = os.path.join(exp_path, version, model_dir + "_finetuned.pt")
         else:
@@ -684,8 +588,8 @@ class TransformersWrapper(ABC):
                                 Defaults to 0.1.
             toks_per_batch: Maximum number of token to consider in a batch.Defaults to 2048.
                             This argument will set the number of sequences in a batch, which
-                            is dynamically computed. Batch size use accumulate_grad_batches to compute
-                            accumulate_grad_batches parameter.
+                            is dynamically computed. Batch size use accumulate_grad_batches
+                            to compute accumulate_grad_batches parameter.
             extra_toks_per_seq: Defaults to 2,
             filter_len : Size of sequence to filter. Defaults to None. (NOT USED)
             accelerator: type of accelerator for mutli-gpu processing (DPP recommanded)
@@ -696,13 +600,17 @@ class TransformersWrapper(ABC):
             logs_name_exp: Name of the experience in the logs.
             checkpoint : Path to a checkpoint file to restore training session.
             save_last_checkpoint: Save last checkpoint and 2 best trainings models
-                                  to restore training session. Take a large amout of time and memory.
+                                  to restore training session. Take a large amout of time
+                                  and memory.
         """
         if isinstance(train_sequences, str):
             train_sequences = load_fasta(train_sequences)
 
-        fit_model = self.model.module if self.multi_gpu else self.model  # type: ignore
-        alphabet = self._get_alphabet_dataloader()
+        # Free resources used by ray before finetuning with Lightning
+        del self._workers
+
+        fit_model = self._language_model.model  # type: ignore
+        alphabet = self._language_model.get_alphabet_dataloader()
 
         extra_toks_per_seq = int(alphabet.prepend_bos) + int(alphabet.append_eos)
         lightning_model = LightningModule(
@@ -725,11 +633,8 @@ class TransformersWrapper(ABC):
             extra_toks_per_seq,
         )
 
-        if torch.cuda.is_available():
-            n_gpus = torch.cuda.device_count()
-        else:
-            log.warning("You try to train a transformers without GPU.")
-            return
+        if self._num_gpus == 0:
+            raise ValueError("You try to train a transformers without GPU.")
 
         logger = CSVLogger(logs_save_dir, name=logs_name_exp)
         checkpoint_callback = None
@@ -746,7 +651,7 @@ class TransformersWrapper(ABC):
             ]
 
         trainer = Trainer(
-            gpus=n_gpus,
+            gpus=self._num_gpus,
             amp_level=amp_level,
             precision=precision,
             accumulate_grad_batches=acc_batch_size // batch_size,
@@ -765,13 +670,20 @@ class TransformersWrapper(ABC):
             rank = os.environ.get("LOCAL_RANK", None)
             rank = int(rank) if rank is not None else None  # type: ignore
             if rank == 0:
-                self.save_model(save_path, lightning_model)
+                save_name = self._save_model(save_path, lightning_model)
         else:
-            self.save_model(save_path, lightning_model)
+            save_name = self._save_model(save_path, lightning_model)
 
-        if self.multi_gpu:
-            self.model = DataParallel(lightning_model.model).to(self._device)
-        else:
-            self.model = lightning_model.model.to(self._device)
+        # Load new model
+        self._language_model._load_model(save_name)
+
+        if self._multi_gpus:
+            # Create ray workers as they have been deleted at the beginning
+            self._workers = [
+                self._ray_cls.remote(model_dir=self._model_dir, device="cuda:0")
+                for _ in range(self._num_gpus)
+            ]
+            # Load new model one ach worker
+            ray.get([worker._load_model.remote(save_name) for worker in self._workers])
 
         log.info("Training completed.")
